# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015, 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.


"""Minimal Flask application example for development.

Create database and tables:

.. code-block:: console

   $ cd examples
   $ flask -a app.py db init
   $ flask -a app.py db create

Create some users:

.. code-block:: console

   $ flask -a app.py users create -e info@invenio-software.org -a
   $ flask -a app.py users create -e reader@invenio-software.org -a
   $ flask -a app.py users create -e editor@invenio-software.org -a
   $ flask -a app.py users create -e admin@invenio-software.org -a

Add a role to a user:

.. code-block:: console

   $ flask -a app.py roles create -n admin
   $ flask -a app.py roles add -u info@invenio-software.org -r admin
   $ flask -a app.py roles add -u admin@invenio-software.org -r admin

Assign some allowed actions to this users:

.. code-block:: console

   $ flask -a app.py access allow open -e editor@invenio-software.org
   $ flask -a app.py access deny open -e info@invenio-software.org
   $ flask -a app.py access allow read -e reader@invenio-software.org
   $ flask -a app.py access allow open -r admin
   $ flask -a app.py access allow read -r admin


Run the development server:

.. code-block:: console

   $ flask -a app.py run

If you are using an action in your app which does not have any role or user
assigned, the action will be allowed to perform by everyone (Anonymous users
included).

"""

from __future__ import absolute_import, print_function

from flask import Flask, g, render_template
from flask.ext.principal import ActionNeed, RoleNeed
from flask_babelex import Babel
from flask_cli import FlaskCLI
from flask_login import current_user
from flask_mail import Mail
from flask_menu import Menu
from invenio_accounts import InvenioAccounts
from invenio_accounts.views import blueprint
from invenio_db import InvenioDB

from invenio_access import InvenioAccess
from invenio_access.permissions import DynamicPermission

# Create Flask application
app = Flask(__name__)
app.secret_key = 'ExampleApp'
FlaskCLI(app)
Babel(app)
Mail(app)
Menu(app)
InvenioDB(app)
InvenioAccounts(app)
app.register_blueprint(blueprint)

access = InvenioAccess(app)

action_open = ActionNeed('open')
access.register_action(action_open)

action_read = ActionNeed('read')
access.register_action(action_read)


@app.route("/")
def index():
    """Basic test view."""
    identity = g.identity
    actions = {}
    for action in access.actions:
        actions[action.value] = DynamicPermission(action).allows(identity)

    if current_user.is_anonymous:
        return render_template("invenio_access/open.html",
                               actions=actions,
                               identity=identity)
    else:
        return render_template("invenio_access/limited.html",
                               message='',
                               actions=actions,
                               identity=identity)


admin_permission = DynamicPermission(RoleNeed('admin'))


@app.route('/role_admin')
@admin_permission.require()
def role_admin():
    """View only allowed to admin role."""
    identity = g.identity
    actions = {}
    for action in access.actions:
        actions[action.value] = DynamicPermission(action).allows(identity)

    message = 'You are opening a page limited to action read'
    return render_template("invenio_access/limited.html",
                           message=message,
                           actions=actions,
                           identity=identity)


open_permission = DynamicPermission(action_open)


@app.route('/action_open')
@open_permission.require()
def action_open():
    """View only allowed to open action."""
    identity = g.identity
    actions = {}
    for action in access.actions:
        actions[action.value] = DynamicPermission(action).allows(identity)

    message = 'You are opening a page limited to action read'
    return render_template("invenio_access/limited.html",
                           message=message,
                           actions=actions,
                           identity=identity)


read_permission = DynamicPermission(action_read)


@app.route('/action_read')
@read_permission.require()
def action_read():
    """View only allowed to open action."""
    identity = g.identity
    actions = {}
    for action in access.actions:
        actions[action.value] = DynamicPermission(action).allows(identity)

    message = 'You are opening a page limited to action read'
    return render_template("invenio_access/limited.html",
                           message=message,
                           actions=actions,
                           identity=identity)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015, 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.


"""Module tests."""

from __future__ import absolute_import, print_function

from flask_principal import ActionNeed, RoleNeed, UserNeed
from invenio_accounts.models import Role, User
from invenio_db import db

from invenio_access import InvenioAccess
from invenio_access.models import ActionRoles, ActionUsers
from invenio_access.permissions import DynamicPermission


class FakeIdentity(object):
    """Fake class to test DynamicPermission."""

    def __init__(self, *provides):
        self.provides = set(provides)


def test_invenio_access_permissions_deny(app):
    """User without any provides can't access to a place limited to user 0"""
    InvenioAccess(app)
    with app.test_request_context():
        permission = DynamicPermission(UserNeed(0))

        fake_identity = FakeIdentity()
        assert not permission.allows(fake_identity)


def test_invenio_access_permission_for_users(app):
    """User can access to an action allowed/denied to the user"""
    InvenioAccess(app)
    with app.test_request_context():
        db.session.begin(nested=True)
        superuser = User(email='superuser@invenio-software.org')
        user_can_all = User(email='all@invenio-software.org')
        user_can_read = User(email='read@invenio-software.org')
        user_can_open = User(email='open@invenio-software.org')

        db.session.add(superuser)
        db.session.add(user_can_all)
        db.session.add(user_can_read)
        db.session.add(user_can_open)

        db.session.add(ActionUsers(action='superuser-access', user=superuser))

        db.session.add(ActionUsers(action='open', user=user_can_all))
        db.session.add(ActionUsers(action='open', user=user_can_open))

        db.session.add(ActionUsers(action='read', user=user_can_all))
        db.session.add(ActionUsers(action='read', user=user_can_read))
        db.session.commit()

        permission_open = DynamicPermission(ActionNeed('open'))
        permission_read = DynamicPermission(ActionNeed('read'))

        identity_superuser = FakeIdentity(UserNeed(superuser.id))
        identity_all = FakeIdentity(UserNeed(user_can_all.id))
        identity_read = FakeIdentity(UserNeed(user_can_read.id))
        identity_open = FakeIdentity(UserNeed(user_can_open.id))

        assert permission_open.allows(identity_superuser)
        assert permission_read.allows(identity_superuser)

        assert permission_open.allows(identity_all)
        assert permission_read.allows(identity_all)

        assert permission_open.allows(identity_open)
        assert not permission_read.allows(identity_open)

        assert not permission_open.allows(identity_read)
        assert permission_read.allows(identity_read)


def test_invenio_access_permission_for_roles(app):
    """User with a role can access to an action allowed to the role"""
    InvenioAccess(app)
    with app.test_request_context():
        superuser_role = Role(name='superuser')
        admin_role = Role(name='admin')
        reader_role = Role(name='reader')
        opener_role = Role(name='opener')

        db.session.add(superuser_role)
        db.session.add(admin_role)
        db.session.add(reader_role)
        db.session.add(opener_role)

        db.session.add(
            ActionRoles(action='superuser-access', role=superuser_role))

        db.session.add(ActionRoles(action='open', role=admin_role))
        db.session.add(ActionRoles(action='open', role=opener_role))

        db.session.add(ActionRoles(action='read', role=admin_role))
        db.session.add(ActionRoles(action='read', role=reader_role))

        db.session.commit()

    with app.app_context():
        permission_open = DynamicPermission(ActionNeed('open'))
        permission_read = DynamicPermission(ActionNeed('read'))

        identity_superuser = FakeIdentity(RoleNeed('superuser'))
        identity_all = FakeIdentity(RoleNeed('admin'))
        identity_read = FakeIdentity(RoleNeed('reader'))
        identity_open = FakeIdentity(RoleNeed('opener'))

        assert permission_open.allows(identity_superuser)
        assert permission_read.allows(identity_superuser)

        assert permission_open.allows(identity_all)
        assert permission_read.allows(identity_all)

        assert permission_open.allows(identity_open)
        assert not permission_read.allows(identity_open)

        assert not permission_open.allows(identity_read)
        assert permission_read.allows(identity_read)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Invenio module for common role based access control."""

from __future__ import absolute_import, print_function

import pkg_resources

from .cli import access as access_cli


class _AccessState(object):
    """Access state storing registered actions."""

    def __init__(self, app, entry_point_group=None, cache=None):
        """Initialize state."""
        self.app = app
        self.actions = {}
        self.cache = cache
        if entry_point_group:
            self.load_entry_point_group(entry_point_group)

    def set_action_cache(self, action_name, data):
        """Store action needs and excludes."""
        if self.cache:
            self.cache.set(
                self.app.config['ACCESS_ACTION_CACHE_PREFIX'] +
                str(action_name), data
            )

    def get_action_cache(self, action_name):
        """Get action needs and excludes from cache."""
        data = None
        if self.cache:
            data = self.cache.get(
                self.app.config['ACCESS_ACTION_CACHE_PREFIX'] +
                str(action_name)
            )
        return data

    def delete_action_cache(self, action_name):
        """Delete action needs and excludes from cache."""
        if self.cache:
            self.cache.delete(
                self.app.config['ACCESS_ACTION_CACHE_PREFIX'] +
                str(action_name)
            )

    def register_action(self, action):
        """Register an action to be showed in the actions list."""
        assert action.value not in self.actions
        self.actions[action.value] = action

    def load_entry_point_group(self, entry_point_group):
        """Load actions from an entry point group."""
        for ep in pkg_resources.iter_entry_points(group=entry_point_group):
            self.register_action(ep.load())


class InvenioAccess(object):
    """Invenio Access extension."""

    def __init__(self, app=None, **kwargs):
        """Extension initialization."""
        if app:
            self._state = self.init_app(app, **kwargs)

    def init_app(self, app, entry_point_group='invenio_access.actions',
                 **kwargs):
        """Flask application initialization."""
        self.init_config(app)
        app.cli.add_command(access_cli, 'access')
        state = _AccessState(app, entry_point_group=entry_point_group,
                             cache=kwargs.get('cache'))
        app.extensions['invenio-access'] = state
        return state

    def init_config(self, app):
        """Initialize configuration."""
        app.config.setdefault('ACCESS_ACTION_CACHE_PREFIX',
                              'DynamicPermission::action::')

    def __getattr__(self, name):
        """Proxy to state object."""
        return getattr(self._state, name, None)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Admin views for managing access to actions."""

from flask import current_app, flash
from flask_admin.actions import action
from flask_admin.contrib.sqla import ModelView
from flask_admin.form.fields import DateTimeField
from flask_login import login_user, logout_user
from werkzeug.local import LocalProxy
from wtforms import SelectField

from .models import ActionRoles, ActionUsers
from .proxies import current_access

_datastore = LocalProxy(lambda: current_app.extensions['security'].datastore)


def _(x):
    """Identity."""
    return x


class ActionUsersView(ModelView):
    """View for managing access to actions by users."""

    can_view_details = True

    list_all = ('user_id', 'user.email', 'action', 'exclude', 'argument')

    column_list = list_all
    column_default_sort = ('user_id', True)

    column_labels = {
        'user.email': _("Email"),
        'user_id': _("User ID"),
    }

    column_filters = list_all

    form_args = dict(
        action=dict(
            choices=LocalProxy(lambda: [
                (action, action) for action in current_access.actions.keys()
            ])
        )
    )
    form_overrides = dict(
        action=SelectField,
    )


class ActionRolesView(ModelView):
    """View for managing access to actions by users with given role."""

    can_view_details = True

    list_all = ('role.name', 'action', 'exclude', 'argument')

    column_list = list_all

    column_filters = \
        columns_sortable_list = \
        columns_searchable_list = \
        list_all

    column_display_all_relations = True

    column_labels = {
        'role.name': _("Role Name"),
    }


action_roles_adminview = {
    'model': ActionRoles,
    'modelview': ActionRolesView,
    'category': _('User Management'),
}

action_users_adminview = {
    'model': ActionUsers,
    'modelview': ActionUsersView,
    'category': _('User Management'),
}

__all__ = ('action_users_adminview', 'action_roles_adminview', )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.


"""Pytest configuration."""

from __future__ import absolute_import, print_function

import os

import pytest
from flask import Flask
from flask.ext.principal import ActionNeed
from flask_babelex import Babel
from flask_cli import FlaskCLI, ScriptInfo
from flask_mail import Mail
from invenio_accounts import InvenioAccounts
from invenio_db import InvenioDB, db

from invenio_access import InvenioAccess
from invenio_access.permissions import ParameterizedActionNeed


@pytest.fixture()
def app(request):
    """Flask application fixture."""
    app = Flask('testapp')
    app.config.update(
        ACCOUNTS_USE_CELERY=False,
        SECRET_KEY="CHANGE_ME",
        SECURITY_PASSWORD_SALT="CHANGE_ME_ALSO",
        SQLALCHEMY_DATABASE_URI=os.environ.get(
            'SQLALCHEMY_DATABASE_URI', 'sqlite:///test.db'),
        TESTING=True,
    )
    FlaskCLI(app)
    Babel(app)
    Mail(app)
    InvenioDB(app)
    InvenioAccounts(app)

    with app.app_context():
        db.create_all()

    def teardown():
        with app.app_context():
            db.drop_all()

    request.addfinalizer(teardown)
    return app


@pytest.fixture()
def script_info(request):
    """Get ScriptInfo object for testing CLI."""
    action_open = ActionNeed('open')
    action_edit = ParameterizedActionNeed('edit', None)

    app_ = app(request)
    ext = InvenioAccess(app_)
    ext.register_action(action_open)
    ext.register_action(action_edit)
    return ScriptInfo(create_app=lambda info: app_)


@pytest.fixture()
def script_info_cli_list(request):
    """Get ScriptInfo object for testing CLI list command."""
    action_open = ActionNeed('open')
    action_edit = ParameterizedActionNeed('edit', None)
    app_ = app(request)
    ext = InvenioAccess(app_)
    ext.register_action(action_open)
    ext.register_action(action_edit)

    return ScriptInfo(create_app=lambda info: app_)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.


"""Minimal Flask application example for admin views.

Create database and tables:

.. code-block:: console

   $ cd examples
   $ flask -a adminapp.py db init
   $ flask -a adminapp.py db create

Create some users:

.. code-block:: console

   $ flask -a adminapp.py users create info@invenio-software.org -a
   $ flask -a adminapp.py users create reader@invenio-software.org -a
   $ flask -a adminapp.py users create editor@invenio-software.org -a
   $ flask -a adminapp.py users create admin@invenio-software.org -a

Run the development server:

.. code-block:: console

   $ flask -a adminapp.py run

"""

from __future__ import absolute_import, print_function

from flask import Flask
from flask_admin import Admin
from flask_babelex import Babel
from flask_cli import FlaskCLI
from flask_mail import Mail
from flask_menu import Menu
from invenio_accounts import InvenioAccounts
from invenio_accounts.views import blueprint
from invenio_admin import InvenioAdmin
from invenio_db import InvenioDB

from invenio_access import InvenioAccess

app = Flask(__name__)
app.secret_key = 'ExampleApp'
FlaskCLI(app)
Babel(app)
Mail(app)
Menu(app)
InvenioDB(app)
InvenioAccounts(app)
access = InvenioAccess(app)

InvenioAdmin(app, permission_factory=lambda x: x,
             view_class_factory=lambda x: x)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015, 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.


"""Module tests."""

from __future__ import absolute_import, print_function

from flask import Flask
from flask_babelex import Babel
from flask_cli import FlaskCLI
from flask_mail import Mail
from flask_principal import ActionNeed
from invenio_accounts import InvenioAccounts
from invenio_db import InvenioDB
from mock import patch
from pkg_resources import EntryPoint

from invenio_access import InvenioAccess, current_access


class MockEntryPoint(EntryPoint):
    """Mocking of entrypoint."""

    def load(self):
        """Mock load entry point."""
        return ActionNeed(self.name)


def _mock_entry_points(group=None):
    """Mocking funtion of entrypoints."""
    data = {
        'invenio_access.actions': [MockEntryPoint('open',
                                                  'demo.actions'),
                                   MockEntryPoint('close',
                                                  'demo.actions')],
    }
    names = data.keys() if group is None else [group]
    for key in names:
        for entry_point in data[key]:
            yield entry_point


def test_version():
    """Test version import."""
    from invenio_access import __version__
    assert __version__


def test_init():
    """Test extension initialization."""
    app = Flask('testapp')
    FlaskCLI(app)
    Babel(app)
    Mail(app)
    InvenioDB(app)
    InvenioAccounts(app)
    ext = InvenioAccess(app)
    assert 'invenio-access' in app.extensions

    app = Flask('testapp')
    FlaskCLI(app)
    Babel(app)
    Mail(app)
    InvenioDB(app)
    InvenioAccounts(app)
    ext = InvenioAccess()
    assert 'invenio-access' not in app.extensions
    ext.init_app(app)
    assert 'invenio-access' in app.extensions


def test_actions():
    """Test if the actions are registered properly."""
    app = Flask('testapp')
    FlaskCLI(app)
    Babel(app)
    Mail(app)
    InvenioDB(app)
    InvenioAccounts(app)
    InvenioAccess(app, entry_point_group=None)
    with app.app_context():
        current_access.register_action(ActionNeed('action_a'))
        assert len(current_access.actions) == 1
        current_access.register_action(ActionNeed('action_b'))
        assert len(current_access.actions) == 2


@patch('pkg_resources.iter_entry_points', _mock_entry_points)
def test_actions_entrypoint():
    """Test if the entrypoint is registering actions properly."""
    app = Flask('testapp')
    FlaskCLI(app)
    ext = InvenioAccess(app)
    assert len(ext.actions) == 2
    assert ActionNeed('open') in ext.actions.values()
    assert ActionNeed('close') in ext.actions.values()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015, 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.


"""Module tests."""

from __future__ import absolute_import, print_function

from click.testing import CliRunner
from flask import g
from flask_principal import ActionNeed
from flask_security.core import _security
from flask_security.utils import login_user
from invenio_accounts.cli import roles_add, roles_create, users_create

from invenio_access.cli import access
from invenio_access.permissions import DynamicPermission, \
    ParameterizedActionNeed


def test_access_cli_allow_action_empty(script_info):
    """Test add action role in access CLI."""
    runner = CliRunner()

    result = runner.invoke(
        access, ['allow', 'open'], obj=script_info)
    assert result.exit_code != 0


def test_access_cli_allow_action_unknown_email(script_info):
    """Test add action role in access CLI."""
    runner = CliRunner()

    result = runner.invoke(
        access, ['allow', 'open', '-e', 'unknown'],
        obj=script_info)
    assert result.exit_code != 0


def test_access_cli_allow_action_unknown_role(script_info):
    """Test add action role in access CLI."""
    runner = CliRunner()

    result = runner.invoke(
        access,
        ['allow', 'open', '-r', 'unknown'],
        obj=script_info)
    assert result.exit_code != 0


def test_access_cli_allow_action_user(script_info):
    """Test add action role in access CLI."""
    runner = CliRunner()

    # User creation
    result = runner.invoke(
        users_create,
        ['a@example.org', '--password', '123456'],
        obj=script_info
    )
    assert result.exit_code == 0
    result = runner.invoke(
        access, ['allow', 'open', '-e', 'a@example.org'], obj=script_info)
    assert result.exit_code == 0


def test_access_cli_allow_action_role(script_info):
    """Test add action role in access CLI."""
    runner = CliRunner()

    # Role creation
    result = runner.invoke(
        roles_create,
        ['open_role'],
        obj=script_info)
    assert result.exit_code == 0

    result = runner.invoke(
        access,
        ['allow', 'open', '-r', 'open_role'],
        obj=script_info
    )
    assert result.exit_code == 0


def test_access_cli_list(script_info_cli_list):
    """Test of list cli command."""
    runner = CliRunner()
    result = runner.invoke(
        access, ['list'],
        obj=script_info_cli_list
    )
    assert result.exit_code == 0
    assert result.output.find("open") != -1


def test_access_matrix(script_info_cli_list):
    """Test of combinations of cli commands."""
    script_info = script_info_cli_list
    runner = CliRunner()

    user_roles = {
        'admin@invenio-software.org': ['admin'],
        'admin-edit@invenio-software.org': ['admin'],
        'open@invenio-software.org': ['opener'],
        'edit@invenio-software.org': ['editor'],
        'info@invenio-software.org': ['opener'],
    }

    action_roles = {
        'open': ['admin', 'opener'],
        'edit': ['admin', 'editor'],
    }

    for role in {role for roles in user_roles.values() for role in roles}:
        # Role creation
        result = runner.invoke(
            roles_create, [role], obj=script_info
        )
        assert result.exit_code == 0

    for email, roles in user_roles.items():
        result = runner.invoke(
            users_create,
            [email, '--password', '123456', '-a'],
            obj=script_info
        )
        assert result.exit_code == 0

        for role in roles:
            # Role creation
            result = runner.invoke(
                roles_add, [email, role], obj=script_info
            )
            assert result.exit_code == 0

    def role_args(roles):
        """Generate role arguments."""
        for role in roles:
            yield '-r'
            yield role

    for action, roles in action_roles.items():

        result = runner.invoke(
            access,
            ['allow', action] + list(role_args(roles)),
            obj=script_info
        )
        assert result.exit_code == 0

    result = runner.invoke(
        access,
        ['deny', 'edit', '-e', 'admin-edit@invenio-software.org'],
        obj=script_info
    )
    assert result.exit_code == 0

    result = runner.invoke(
        access,
        ['allow', 'edit', '-a', '1', '-e', 'info@invenio-software.org'],
        obj=script_info
    )
    assert result.exit_code == 0

    permission_open = DynamicPermission(ActionNeed('open'))
    permission_edit = DynamicPermission(ActionNeed('edit'))

    permission_edit_1 = DynamicPermission(ParameterizedActionNeed('edit', 1))
    permission_edit_2 = DynamicPermission(ParameterizedActionNeed('edit', 2))

    user_permissions = {
        'admin@invenio-software.org': {
            True: [
                permission_open,
                permission_edit,
                permission_edit_1,
                permission_edit_2,
            ],
        },
        'admin-edit@invenio-software.org': {
            True: [permission_open],
            False: [
                permission_edit,
                permission_edit_1,
                permission_edit_2,
            ],
        },
        'open@invenio-software.org': {
            True: [permission_open],
            False: [
                permission_edit,
                permission_edit_1,
                permission_edit_2,
            ],
        },
        'edit@invenio-software.org': {
            True: [
                permission_edit,
                permission_edit_1,
                permission_edit_2,
            ],
            False: [permission_open],
        },
        'info@invenio-software.org': {
            True: [
                permission_open,
                permission_edit_1,
            ],
            False: [
                permission_edit,
                permission_edit_2,
            ],
        },
    }

    for email, permissions in user_permissions.items():
        with script_info.create_app(None).test_request_context():
            user = _security.datastore.find_user(email=email)
            login_user(user)
            identity = g.identity

            for can, actions in permissions.items():
                for action in actions:
                    assert action.allows(identity) == can

    result = runner.invoke(
        access,
        ['remove', 'edit'],
        obj=script_info
    )
    assert result.exit_code == 2

    result = runner.invoke(
        access,
        ['show', '-e', 'admin-edit@invenio-software.org'],
        obj=script_info
    )
    assert result.exit_code == 0
    assert 'user:admin-edit@invenio-software.org:edit::deny' in result.output

    result = runner.invoke(
        access,
        ['show', '-r', 'editor'],
        obj=script_info
    )
    assert 'role:editor:edit::allow\n' == result.output
    assert result.exit_code == 0

    #
    # Remove all permissions.
    #
    for action, roles in action_roles.items():

        result = runner.invoke(
            access,
            ['remove', action] + list(role_args(roles)),
            obj=script_info
        )
        assert result.exit_code == 0

    result = runner.invoke(
        access,
        ['remove', 'edit', '-e', 'admin-edit@invenio-software.org'],
        obj=script_info
    )
    assert result.exit_code == 0

    result = runner.invoke(
        access,
        ['remove', 'edit', '-a', '1', '-e', 'info@invenio-software.org'],
        obj=script_info
    )
    assert result.exit_code == 0

    # All authorizations should be removed.
    result = runner.invoke(
        access,
        ['show', '-r', 'admin-edit@invenio-software.org'],
        obj=script_info
    )
    assert result.exit_code == 0
    assert result.output == ''



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015, 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Invenio module for advanced permission control.

Dynamic permission control is a core feature of *Invenio-Access*
package.  It allows administrator to fine-tune access to various
actions with parameters to a specific user or role.
"""

from collections import namedtuple
from functools import partial
from itertools import chain

from flask_principal import ActionNeed, Permission

from .models import ActionRoles, ActionUsers
from .proxies import current_access

_Need = namedtuple('Need', ['method', 'value', 'argument'])

ParameterizedActionNeed = partial(_Need, 'action')

ParameterizedActionNeed.__doc__ = \
    """A need with the method preset to `"action"` with parameter.

    If it is called with `argument=None` then this need is equivalent
    to ``ActionNeed``.
    """

superuser_access = ActionNeed('superuser-access')
"""SuperUser access allows access to everything on Invenio.

DynamicPermissions handles allowing access to SuperUser.
"""


class _P(namedtuple('Permission', ['needs', 'excludes'])):
    """Helper for simple permission updates."""

    def update(self, permission):
        """In-place update of permissions."""
        self.needs.update(permission.needs)
        self.excludes.update(permission.excludes)


class DynamicPermission(Permission):
    """Utility class providing lazy loading of permissions.

    .. warning::

        If ``ActionNeed`` or ``ParameterizedActionNeed`` is not allowed or
        restricted to any user or role then it is **ALLOWED** to anybody.
        This is a major diference to standard ``Permission`` class.

    """

    def __init__(self, *needs):
        """Default constructor."""
        self._permissions = None
        self.explicit_needs = set(needs)
        self.explicit_needs.add(superuser_access)

    def _load_permissions(self):
        """Load permissions associated to actions."""
        result = _P(needs=set(), excludes=set())

        for explicit_need in self.explicit_needs:
            if explicit_need.method == 'action':
                action = current_access.get_action_cache(
                    explicit_need.value
                )
                if action is None:
                    action = _P(needs=set(), excludes=set())

                    actionsusers = ActionUsers.query_by_action(
                        explicit_need
                    ).all()

                    actionsroles = ActionRoles.query_by_action(
                        explicit_need
                    ).join(
                        ActionRoles.role
                    ).all()

                    for db_action in chain(actionsusers, actionsroles):
                        if db_action.exclude:
                            action.excludes.add(db_action.need)
                        else:
                            action.needs.add(db_action.need)

                    current_access.set_action_cache(
                        explicit_need.value, action
                    )
                # in-place update of results
                result.update(action)
            else:
                result.needs.add(explicit_need)
        self._permissions = result

    @property
    def needs(self):
        """Return allowed permissions from database.

        If there is no role or user assigned to an action, everyone is allowed
        to perform that action.
        """
        if not self._permissions:
            self._load_permissions()
        return self._permissions.needs

    @property
    def excludes(self):
        """Return denied permissions from database.

        If there is no role or user assigned to an action, everyone is allowed
        to perform that action.
        """
        if not self._permissions:
            self._load_permissions()  # pragma: no cover
        return self._permissions.excludes



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015, 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Database models for access module."""

from __future__ import absolute_import, print_function

from flask_principal import RoleNeed, UserNeed
from invenio_accounts.models import Role, User
from invenio_db import db
from sqlalchemy import UniqueConstraint
from sqlalchemy.event import listen

from .proxies import current_access


class ActionNeedMixin(object):
    """Define common attributes for Action needs."""

    id = db.Column(db.Integer, autoincrement=True, primary_key=True)
    """Primary key to support nullable fields."""

    action = db.Column(db.String(80), index=True)
    """Name of the action."""

    exclude = db.Column(db.Boolean(), nullable=False,
                        default=False, server_default="0")
    """Deny associated objects."""

    argument = db.Column(db.String(255), nullable=True, index=True)
    """String value of action argument."""

    @classmethod
    def create(cls, action, **kwargs):
        """Create new database instance from action need."""
        assert action.method == 'action'
        argument = kwargs.pop('argument', None) or getattr(
            action, 'argument', None)
        return cls(
            action=action.value,
            argument=argument,
            **kwargs
        )

    @classmethod
    def allow(cls, action, **kwargs):
        """Allow given action need."""
        return cls.create(action, exclude=False, **kwargs)

    @classmethod
    def deny(cls, action, **kwargs):
        """Deny usage of given action need."""
        return cls.create(action, exclude=True, **kwargs)

    @classmethod
    def query_by_action(cls, action, argument=None):
        """Prepare query object with filtered action."""
        query = cls.query.filter_by(action=action.value)
        argument = argument or getattr(action, 'argument', None)
        if argument is not None:
            query = query.filter(db.or_(
                cls.argument == str(argument),
                cls.argument.is_(None),
            ))
        else:
            query = query.filter(cls.argument.is_(None))
        return query

    @property
    def need(self):
        """Abstract need."""
        raise NotImplemented()  # pragma: no cover


class ActionUsers(ActionNeedMixin, db.Model):
    """ActionRoles data model.

    It relates an allowed action with a user.
    """

    __tablename__ = 'access_actionsusers'

    __table_args__ = (UniqueConstraint(
        'action', 'exclude', 'argument', 'user_id',
        name='access_actionsusers_unique'),
    )

    user_id = db.Column(db.Integer(), db.ForeignKey(User.id))

    user = db.relationship("User")

    @property
    def need(self):
        """Return UserNeed instance."""
        return UserNeed(self.user_id)


class ActionRoles(ActionNeedMixin, db.Model):
    """ActionRoles data model.

    It relates an allowed action with a role.
    """

    __tablename__ = 'access_actionsroles'

    __table_args__ = (UniqueConstraint(
        'action', 'exclude', 'argument', 'role_id',
        name='access_actionsroles_unique'),
    )

    role_id = db.Column(db.Integer(), db.ForeignKey(Role.id), nullable=False)

    role = db.relationship("Role")

    @property
    def need(self):
        """Return RoleNeed instance."""
        return RoleNeed(self.role.name)


def removed_or_inserted_action(mapper, connection, target):
    """Remove the action from cache when an item is inserted or deleted."""
    current_access.delete_action_cache(target.action)


def changed_action_action(target, value, oldvalue, initiator):
    """Remove the action from cache when an action is updated.

    Remove the action from cache when ActionRoles.action or ActionUsers.action
    is updated.
    """
    current_access.delete_action_cache(value)
    current_access.delete_action_cache(oldvalue)


def changed_owner_action(target, value, oldvalue, initiator):
    """Remove the action from cache when the owner is updated.

    Remove the action from cache when ActionRoles.role or ActionUsers.user
    is updated.
    """
    current_access.delete_action_cache(target.action)


listen(ActionUsers, 'after_insert', removed_or_inserted_action)
listen(ActionUsers, 'after_delete', removed_or_inserted_action)
listen(ActionUsers.action, 'set', changed_action_action)
listen(ActionUsers.user, 'set', changed_owner_action)

listen(ActionRoles, 'after_insert', removed_or_inserted_action)
listen(ActionRoles, 'after_delete', removed_or_inserted_action)
listen(ActionRoles.action, 'set', changed_action_action)
listen(ActionRoles.role, 'set', changed_owner_action)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Command Line Interface for access."""

from __future__ import absolute_import, print_function

import click
from flask import current_app
from flask_cli import with_appcontext
from invenio_accounts.models import Role, User
from invenio_db import db
from werkzeug.local import LocalProxy

from .models import ActionRoles, ActionUsers

_current_actions = LocalProxy(
    lambda: current_app.extensions['invenio-access'].actions
)
"""Helper proxy to registered actions."""


def _store_action(action, email, role, exclude=False, argument=None):
    """Store actions."""
    assert action in _current_actions
    action_need = _current_actions[action]

    if not email and not role:
        raise click.UsageError('You haven\'t specified any user or role.')

    #
    # Assign users
    #
    action_factory = ActionUsers.deny if exclude else ActionUsers.allow

    for email_item in email:
        user = User.query.filter(User.email == email_item).first()
        if not user:
            raise click.BadParameter('User with email \'%s\' not found.',
                                     email_item)

        db.session.add(action_factory(
            action_need, argument=argument, user=user
        ))

    #
    # Assign roles
    #
    action_factory = ActionRoles.deny if exclude else ActionRoles.allow

    for role_item in role:
        role = Role.query.filter(Role.name == role_item).first()
        if not role:
            raise click.BadParameter('Role with name \'%s\' not found.',
                                     role_item)
        db.session.add(action_factory(
            action_need, argument=argument, role=role
        ))

    db.session.commit()


def _remove_action(action, email, role, argument=None):
    """Store actions."""
    assert action in _current_actions
    action_need = _current_actions[action]

    if not email and not role:
        raise click.UsageError('You haven\'t specified any user or role.')

    if email:
        ActionUsers.query_by_action(action_need, argument=argument).filter(
            ActionUsers.user.has(User.email.in_(email))
        ).delete(synchronize_session=False)

    if role:
        ActionRoles.query_by_action(action_need, argument=argument).filter(
            ActionRoles.role.has(Role.name.in_(role))
        ).delete(synchronize_session=False)

    db.session.commit()


option_argument = click.option(
    '-a', '--argument', default=None, metavar='VALUE',
    help='Value for parameterized action.'
)

option_email = click.option(
    '-e', '--email', multiple=True, default=[], metavar='EMAIL',
    help='User email address(es).'
)
option_role = click.option(
    '-r', '--role', multiple=True, default=[], metavar='ROLE',
    help='Role name(s).'
)


#
# Access commands
#
@click.group()
def access():
    """Account commands."""


@access.command(name='allow')
@click.argument('action')
@option_argument
@option_email
@option_role
@with_appcontext
def allow_action(action, argument, email, role):
    """Allow actions."""
    _store_action(action, email, role, exclude=False,
                  argument=argument)


@access.command(name='deny')
@click.argument('action')
@option_argument
@option_email
@option_role
@with_appcontext
def deny_action(action, argument, email, role):
    """Deny actions."""
    _store_action(action, email, role, exclude=True,
                  argument=argument)


@access.command(name='remove')
@click.argument('action')
@option_argument
@option_email
@option_role
@with_appcontext
def remove_action(action, argument, email, role):
    """Remove existing action authorization.

    It is possible to specify multible emails and/or roles that
    should be unassigned from the given action.
    """
    _remove_action(action, email, role, argument=argument)


@access.command(name='list')
@with_appcontext
def list_actions():
    """List all registered actions."""
    for name, action in _current_actions.items():
        click.echo('{0}:{1}'.format(
            name, '*' if hasattr(action, 'argument') else ''
        ))


@access.command(name='show')
@option_email
@option_role
@with_appcontext
def show_actions(email, role):
    """Show all assigned actions."""
    if email:
        actions = ActionUsers.query.join(ActionUsers.user).filter(
            User.email.in_(email)
        ).all()
        for action in actions:
            click.secho('user:{0}:{1}:{2}:{3}'.format(
                action.user.email,
                action.action,
                '' if action.argument is None else action.argument,
                'deny' if action.exclude else 'allow',
            ), fg='red' if action.exclude else 'green')

    if role:
        actions = ActionRoles.query.filter(
            Role.name.in_(role)
        ).join(ActionRoles.role).all()
        for action in actions:
            click.secho('role:{0}:{1}:{2}:{3}'.format(
                action.role.name,
                action.action,
                '' if action.argument is None else action.argument,
                'deny' if action.exclude else 'allow',
            ), fg='red' if action.exclude else 'green')



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Invenio module for common role based access control.

It aims to make a process of managing access rigths quick and easy while
preventing unwanted visitors performing restricted actions due to system
misconfiguration.

Access module in three points:

- parametrized action needs
- arbitrary combinations of rules for allowing/denying users and/or roles
- support lazy loading of permissions at runtime

Initialization
--------------

First create a Flask application (Flask-CLI is not needed for Flask
version 1.0+):

>>> from flask import Flask
>>> from flask_cli import FlaskCLI
>>> app = Flask('myapp')
>>> app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite://'
>>> ext_cli = FlaskCLI(app)

You initialize Access like a normal Flask extension, however Invenio-Access is
dependent on Invenio-DB and Invenio-Accounts so first you need to initialize
these extensions:

>>> from invenio_db import InvenioDB
>>> from invenio_accounts import InvenioAccounts
>>> from invenio_access  import InvenioAccess
>>> ext_db = InvenioDB(app)
>>> ext_accounts = InvenioAccounts(app)
>>> ext_access = InvenioAccess(app)

In order for the following examples to work, you need to work within an
Flask application context so let's push one:

>>> ctx = app.app_context()
>>> ctx.push()

Also, for the examples to work we need to create the database and tables (note,
in this example we use an in-memory SQLite database):

>>> from invenio_db import db
>>> db.create_all()

Dynamic Permissions
-------------------

Dynamic permissions represent needs associated with ``ActionNeed`` or
``ParametrizedActionNeed`` or directly the needs which method is different from
`'action'`. This consequently means that if the action is not attached to
any user or role then it is **allowed** by default.

The permissions are loaded from the database the first time you perform
```needs``` or ```excludes``` in the DynamicPermission object and they are
stored in the instance of DynamicPermission affected to be used in future
calls.

Action Needs
~~~~~~~~~~~~

You can start by creating new actions either with or without parameters:

>>> from flask_principal import ActionNeed
>>> from invenio_access import ParameterizedActionNeed
>>> action_index = ActionNeed('index')
>>> action_view_all = ParameterizedActionNeed('view', None)
>>> action_view_1 = ParameterizedActionNeed('view', '1')

If no role or user is attached anybody can perform the action. You can verify
this behavior using following commands.

>>> from flask_principal import AnonymousIdentity
>>> from invenio_access import DynamicPermission, ParameterizedActionNeed
>>> permission_index = DynamicPermission(action_index)
>>> permission_view_all = DynamicPermission(action_view_all)
>>> permission_view_1 = DynamicPermission(action_view_1)
>>> anonymous_identity = AnonymousIdentity()
>>> anonymous_identity.can(permission_index)
True
>>> anonymous_identity.can(permission_view_all)
True
>>> anonymous_identity.can(permission_view_1)
True

Assign actions
~~~~~~~~~~~~~~

You can either assign actions to user or roles. First, you need to have
some users and roles in your database.

>>> from invenio_db import db
>>> from invenio_accounts.models import User, Role
>>> admin = User(email='admin@invenio-software.org')
>>> student = User(email='student@invenio-software.org')
>>> db.session.begin(nested=True)
<sqlalchemy.orm.session.SessionTransaction object ...
>>> db.session.add(admin)
>>> db.session.add(student)
>>> db.session.commit()

With users created you can allow and deny actions:

>>> from invenio_access.models import ActionUsers
>>> db.session.begin(nested=True)
<sqlalchemy.orm.session.SessionTransaction object ...
>>> db.session.add(ActionUsers.allow(action_index, user=admin))
>>> db.session.commit()
>>> permission_index = DynamicPermission(action_index)
>>> anonymous_identity.can(permission_index)
False
>>> from flask_principal import Identity, RoleNeed, UserNeed
>>> admin_identity = Identity(admin.id)
>>> admin_identity.provides.add(UserNeed(admin.id))
>>> admin_identity.can(permission_index)
True

"""

from __future__ import absolute_import, print_function

from .ext import InvenioAccess
from .permissions import DynamicPermission, ParameterizedActionNeed
from .proxies import current_access
from .version import __version__

__all__ = (
    '__version__',
    'current_access',
    'DynamicPermission',
    'InvenioAccess',
    'ParameterizedActionNeed',
)
